/////////////////////////////////////////////////////////////////////////////
///
/// @file morsecon.h
///
/// Connections between Morse sets.
/// This file contains the definition of two auxiliary classes
/// for recording connections between Morse sets in a Morse decomposition.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2008 by Pawel Pilarczyk.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 11, 2008. Last revision: February 11, 2008.


#ifndef _CMGRAPHS_MORSECON_H_
#define _CMGRAPHS_MORSECON_H_


// include selected header files from the CHomP library
#include "chomp/system/config.h"
#include "chomp/struct/multitab.h"

// include local header files
#include "config.h"
#include "typedefs.h"
#include "morsedec.h"


// --------------------------------------------------
// ---- Connection Recorders for inclusionGraph -----
// --------------------------------------------------

/// A simple class for storing connections in an array that uses
/// the "multitable" class.
class TConnTable
{
public:
	/// The constructor.
	TConnTable (int _n, chomp::homology::multitable
		<chomp::homology::multitable<int> > &_connections,
		chomp::homology::multitable<int> &_conncount,
		const int *_numTranslate);

	/// Adds an element v at the connection from source to target.
	void operator () (int source, int target, int v);

private:
	/// The number of vertices between the connections are recorded.
	int n;

	/// The tables of connectiong orbits; those for v -> w are stored
	/// at the index n * v + w.
	chomp::homology::multitable <chomp::homology::multitable<int> >
		&connections;

	/// The numbers of elements in each connecting orbit.
	chomp::homology::multitable<int> &conncount;

	/// The translation table for vertex numbers in the connections.
	const int *numTranslate;

}; /* class TConnTable */

inline TConnTable::TConnTable (int _n,
	chomp::homology::multitable<chomp::homology::multitable<int> >
	&_connections, chomp::homology::multitable<int> &_conncount,
	const int *_numTranslate):
	n (_n), connections (_connections), conncount (_conncount),
	numTranslate (_numTranslate)
{
	int maxcount = n * n;
	for (int i = 0; i < maxcount; ++ i)
		conncount [i] = 0;
	return;
} /* TConnTable::TConnTable */

inline void TConnTable::operator () (int source, int target, int v)
{
	int pos = n * source + target;
	connections [pos] [conncount [pos] ++] = numTranslate [v];
	return;
} /* TConnTable::operator () */

// --------------------------------------------------

/// A simple class for storing connections in an array that uses
/// the "multitable" class.
class TConnMorse
{
public:
	/// The constructor.
	TConnMorse (theMorseDecompositionType &_morseDec,
		const spcCubes &_X, const int *_numTranslate);

	/// Adds an element v at the connection from source to target.
	void operator () (int source, int target, int v);

private:
	/// The Morse decomposition in which to record the connections.
	theMorseDecompositionType &morseDec;

	/// The set which contains cubes to add.
	const spcCubes &X;

	/// The translation table for vertex numbers in the connections.
	const int *numTranslate;

}; /* class TConnMorse */

inline TConnMorse::TConnMorse (theMorseDecompositionType &_morseDec,
	const spcCubes &_X, const int *_numTranslate):
	morseDec (_morseDec), X (_X), numTranslate (_numTranslate)
{
	return;
} /* TConnMorse::TConnMorse */

inline void TConnMorse::operator () (int source, int target, int v)
{
	morseDec. addconn (source, target, X [numTranslate [v]]);
	return;
} /* TConnMorse::operator () */


#endif // _CMGRAPHS_MORSECON_H_

